/* $Id$ -*- C++ -*-
 */

#include <cassert>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>
#include <cmath>

enum {NA=-999};

class Angle {
  double _a;
  Angle& fix() {if (_a != NA) {while (_a<0) _a+=360; while (_a>360) _a-=360;} return *this;}
public:
  Angle() : _a(NA) {}
  Angle(double p) : _a(p) {fix();}
  Angle(const Angle& p) : _a(p._a) {fix();}
  std::istream& read(std::istream& in) {in >> _a; fix(); return in;}
  Angle& operator+=(double x) {_a += x; return fix();}
  Angle& operator-=(double x) {_a -= x; return fix();}
  Angle& operator*=(double x) {_a *= x; return fix();}
  Angle& operator+=(Angle x) {_a += x._a; return fix();}
  Angle& operator-=(Angle x) {_a -= x._a; return fix();}
  Angle& operator*=(Angle x) {_a *= x._a; return fix();}
  operator double()const {return _a;}
};
std::istream& operator>>(std::istream& in,  Angle& r) {return r.read(in);}

class Radial {
public:
  Angle _angle;
  double _haat;
  double _antenna_rel_field;
  double _antenna_rel_db;
  double _power;
  double _p60;
  double _i100;
  double _i54;
  double _i40;
public:
  Radial();
  Radial(const Radial&);
  Radial(std::istream&);
  operator bool()const {return _angle != NA;}
  std::istream& read(std::istream&);
  std::ostream& write(std::ostream&)const;
  bool operator<(const Radial& x)const {return _angle < x._angle;}
  bool operator==(const Radial& x)const {return _angle == x._angle;}
  Radial& operator+=(const Radial& x);
  Radial& operator-=(const Radial& x);
  Radial& operator*=(double x);
};
std::istream& operator>>(std::istream& in,  Radial& r) {return r.read(in);}
std::ostream& operator<<(std::ostream& out, const Radial& r) {return r.write(out);}


class Station {
public:
  std::string _comment;
  std::string _location;
  double _latitude;  // radians
  double _longitude; // radians
  double _power;
  int _channel;
  std::vector<Radial> _radials;
public:
  typedef std::vector<Radial>::const_iterator const_iterator;
  typedef std::vector<Radial>::iterator iterator;

  Station(std::istream& is) {read(is);}
  void insert(Radial& r) {_radials.push_back(r);}
  std::istream& read(std::istream&);
  std::ostream& write(std::ostream&)const;
  iterator begin() {return _radials.begin();}
  iterator end() {return _radials.end();}
  Radial operator[](double index);
};

//------------------------------------------------------------
inline Radial::Radial()
  :_angle(NA),
   _haat(NA),
   _antenna_rel_field(NA),
   _antenna_rel_db(NA),
   _power(NA),
   _p60(NA),
   _i100(NA),
   _i54(NA),
   _i40(NA)
{
}
//------------------------------------------------------------
inline Radial::Radial(std::istream& in)
  :_angle(NA),
   _haat(NA),
   _antenna_rel_field(NA),
   _antenna_rel_db(NA),
   _power(NA),
   _p60(NA),
   _i100(NA),
   _i54(NA),
   _i40(NA)
{
  read(in);
}
//------------------------------------------------------------
inline Radial::Radial(const Radial & p)
  :_angle(p._angle),
   _haat(p._haat),
   _antenna_rel_field(p._antenna_rel_field),
   _antenna_rel_db(p._antenna_rel_db),
   _power(p._power),
   _p60(p._p60),
   _i100(p._i100),
   _i54(p._i54),
   _i40(p._i40)
{
}
//------------------------------------------------------------
inline std::istream& Radial::read(std::istream& in)
{
  std::string buffer;
  getline(in, buffer);
  std::istringstream is(buffer);
  is >> _angle >> _haat >> _antenna_rel_field >> _antenna_rel_db >> _power
     >> _p60 >> _i40 >> _i54 >> _i100;
  return in;
}
//------------------------------------------------------------
inline std::ostream& Radial::write(std::ostream& out)const
{
  out.precision(4);
  out << _angle
      << '\t' << _haat;
  out.precision(3);
  out << '\t' << _antenna_rel_field
      << '\t' << _antenna_rel_db
      << '\t' << _power;
  out.precision(4);
  out << '\t' << _p60
      << '\t' << _i40
      << '\t' << _i54
      << '\t' << _i100
      << '\n';
  return out;
}
//------------------------------------------------------------
//------------------------------------------------------------
inline std::istream& Station::read(std::istream& in)
{
  getline(in, _comment);
  getline(in, _location);
  std::string junk;
  getline(in, junk);

  std::istringstream is(_location);
  {
    double degrees = 0;
    double minutes = 0;
    double seconds = 0;
    std::string direction;
    is >> degrees >> minutes >> seconds >> direction;
    minutes += seconds / 60;
    degrees += minutes / 60;
    if (direction == "S") {
      degrees = -degrees;
    }
    _latitude = degrees * M_PI / 180;
  }
  {
    double degrees = 0;
    double minutes = 0;
    double seconds = 0;
    std::string direction;
    is >> degrees >> minutes >> seconds >> direction;
    minutes += seconds / 60;
    degrees += minutes / 60;
    if (direction == "W") {
      degrees = -degrees;
    }
    _longitude = degrees * M_PI / 180;
  }
  is >> _power >> _channel;

  for (;;) {
    Radial r(in);
    if (!r) break;
    insert(r);
  }
}
//------------------------------------------------------------
inline std::ostream& Station::write(std::ostream& out)const
{
  out << _comment << '\n';

  {
    double degrees = std::abs(_latitude * 180 / M_PI);
    double d = floor(degrees);
    double minutes = (degrees - d) * 60;
    double m = floor(minutes);
    double s = (minutes - m) * 60;
    out << d << '\t' << m << '\t' << s << ((_latitude >= 0) ? "\tN\t " : "\tS\t");
  }
  {
    double degrees = std::abs(_longitude * 180 / M_PI);
    double d = floor(degrees);
    double minutes = (degrees - d) * 60;
    double m = floor(minutes);
    double s = (minutes - m) * 60;
    out << d << '\t' << m << '\t' << s << ((_longitude >= 0) ? "\tE\t " : "\tW\t");
  }
  out << _power << '\t' << _channel << '\n';

  out << "bearing\thaat\tfield\tdb\tpower\tp60\ti40\ti54\ti100\n";

  for (const_iterator ii = _radials.begin(); ii != _radials.end(); ++ii) {
    ii->write(out);
  }
  return out;
}
//------------------------------------------------------------
Radial& Radial::operator+=(const Radial& p)
{
  _angle += p._angle;
  _haat += p._haat;
  _antenna_rel_field += p._antenna_rel_field;
  _antenna_rel_db += p._antenna_rel_db;
  _power += p._power;
  _p60 += p._p60;
  _i100 += p._i100;
  _i54 += p._i54;
  _i40 += p._i40;
  return *this;
}
//------------------------------------------------------------
Radial& Radial::operator-=(const Radial& p)
{
  _angle -= p._angle;
  _haat -= p._haat;
  _antenna_rel_field -= p._antenna_rel_field;
  _antenna_rel_db -= p._antenna_rel_db;
  _power -= p._power;
  _p60 -= p._p60;
  _i100 -= p._i100;
  _i54 -= p._i54;
  _i40 -= p._i40;
  return *this;
}
//------------------------------------------------------------
Radial& Radial::operator*=(double x)
{
  _angle *= x;
  _haat *= x;
  _antenna_rel_field *= x;
  _antenna_rel_db *= x;
  _power *= x;
  _p60 *= x;
  _i100 *= x;
  _i54 *= x;
  _i40 *= x;
}
//------------------------------------------------------------
Radial operator+(Radial a, const Radial& b)
{
  return a += b;
}
//------------------------------------------------------------
Radial operator-(Radial a, const Radial& b)
{
  return a -= b;
}
//------------------------------------------------------------
Radial operator*(Radial a, double b)
{
  return a *= b;
}
//------------------------------------------------------------
Radial Station::operator[](double index)
{
  Radial idx;
  idx._angle = index;
  const_iterator upper = upper_bound(begin(), end(), idx);
  const_iterator lower = upper-1;
  if (upper == end()) {
    upper = begin();
  }
  double split = (index - lower->_angle) / (upper->_angle - lower->_angle);
  if (split < 0) {
    split = (index - lower->_angle) / (upper->_angle + 360 - lower->_angle);
  }

  return Radial(*lower + (*upper - *lower) * split);
}
//------------------------------------------------------------
//------------------------------------------------------------
